#ifndef HASH_TABLE_HPP_
#define HASH_TABLE_HPP_

#include <list>
#include <vector>

template<typename Value>
struct HashNode
{
  size_t hash;
  Value data;

  HashNode() = delete;
  HashNode(HashNode const&) = delete;
  HashNode(HashNode&&) = delete;
  HashNode& operator=(HashNode const&) = delete;
  HashNode& operator=(HashNode&&) = delete;

  HashNode (size_t hash, Value const& data)
    : hash (hash), data (data)
  {
  }
  HashNode (size_t&& hash, Value&& data)
    : hash (std::move (hash)), data (std::move (data))
  {
  }
};

template<typename Value, typename Hash, typename KeyEqual>
class HashTable;

template<typename Value, typename Hash, typename KeyEqual>
class HashTableIterator
{
public:
  using value_type = Value;
  using difference_type = std::ptrdiff_t;
  using pointer = value_type const*;
  using reference = value_type const&;
  using iterator_category = std::forward_iterator_tag;
private:

  using table = HashTable<Value, Hash, KeyEqual>;
  using local_iterator = typename table::local_iterator;

  size_t
  seek_first_bucket(table const* table) noexcept
  {
    size_t bucket = 0;
    while (bucket != m_table->bucket_count() && m_table->bucket_size(bucket) == 0)
    {
      ++bucket;
    }
    return bucket;
  }

public:
  // begin
  HashTableIterator (table const* table)
    : m_table (table), m_bucket (seek_first_bucket(table)), m_local (m_table->begin (m_bucket))
  {
  }

  // end
  HashTableIterator (table const* table, size_t)
    : m_table (table), m_bucket (m_table->bucket_count ()), m_local ()
  {
  }

  // arbitrary
  HashTableIterator (table const* table, size_t bucket, local_iterator it)
    : m_table (table), m_bucket (bucket), m_local (it)
  {
  }

  Value const& operator* () const
  {
    return m_local->data;
  }

  Value const* operator-> () const
  {
    return std::addressof (m_local->data);
  }

  table const*
  hash_table() const
  {
    return m_table;
  }

  size_t
  bucket () const
  {
    return m_bucket;
  }

  local_iterator
  local () const
  {
    return m_local;
  }

  HashTableIterator&
  operator++ ()
  {
    // [4] TODO
    //
    // advance the "local iterator"
    // while it's equal to end of the bucket:
    //    advance the bucket
    //    if the current bucket is the total bucket count
    //       update local to a default-constructed iterator
    //    else:
    //       update local to the beginning of the current bucket

  }

  HashTableIterator
  operator++ (int)
  {
    HashTableIterator copy (*this);
    ++(*this);
    return copy;
  }

  friend bool
  operator== (HashTableIterator const& i, HashTableIterator const& j)
  {
    // [4] TODO
    //
    // make sure we reference the same table
    // make sure we reference the same bucket
    // if the bucket is the tables, bucket count, return true
    // if the local iterators match, return true
    // in ANY other case, the iterators are not the same

  }

  friend bool
  operator!= (HashTableIterator const& i, HashTableIterator const& j)
  {
    return !(i == j);
  }

private:
  table const* m_table;
  size_t m_bucket;
  local_iterator m_local;

  friend class HashTable<Value, Hash, KeyEqual>;
};

template<typename Value, typename Hash, typename KeyEqual>
class HashTable
{
  using Entry = HashNode<Value>;
  using Bucket = std::list<Entry>;
  using Table = std::vector<Bucket>;

  static constexpr float DEFAULT_LOAD_FACTOR = 0.75f;

public:
  static constexpr size_t DEFAULT_BUCKET_COUNT = 13;

  using iterator = HashTableIterator<Value, Hash, KeyEqual>;
  using local_iterator = typename Bucket::const_iterator;

  HashTable ()
    : m_buckets (DEFAULT_BUCKET_COUNT),
      m_hash (),
      m_key_equal (),
      m_size (0)
  {
  }

  HashTable (size_t bucket_count, Hash const& hash, KeyEqual const& key_equal)
    : m_buckets (bucket_count)
    , m_hash (hash)
    , m_key_equal (key_equal)
    , m_size (0)
  {
  }

  // -------------------- Iterators --------------------

  iterator
  begin () const noexcept
  {
    // [1] TODO

  }

  iterator
  cbegin () const noexcept
  {
    return begin ();
  }

  iterator
  end () const noexcept
  {
    // [1] TODO

  }

  iterator
  cend () const noexcept
  {
    return end ();
  }

  local_iterator
  begin (size_t n) const noexcept
  {
    return m_buckets[n].begin ();
  }

  local_iterator
  cbegin (size_t n) const noexcept
  {
    return begin (n);
  }

  local_iterator
  end (size_t n) const noexcept
  {
    return m_buckets[n].end ();
  }

  local_iterator
  cend (size_t n) const noexcept
  {
    return end (n);
  }

  // -------------------- Mutators --------------------

  void
  clear ()
  {
    // [4] TODO
    //
    // clear the buckets -- number of buckets should stay the same
    // reset size to zero

  }

  std::pair<iterator, bool>
  insert_unique (Value const& value)
  {
    // [9] TODO
    //
    // 1. hash the value, and find the bucket
    // 2. iterate through the bucket and ask if any items have an equal key
    //      if any match, return a pair corresponding to an iterator + false
    //
    // 3. increment the size
    // 4. if load_factor() > max_load_factor()
    //     rehash -- the bucket count should grow by 2 * n + 1
    //     determine the new bucket for the value -- DO NOT RECOMPUTE THE HASH
    // 5. emplace a new entry (hash-value) at the end of the bucket
    // 6. use the iterator returned from emplace to make a hash table iterator
    // 7. return iterator + true pair

  }

  iterator
  insert (Value const& value)
  {
    // [9] TODO
    //
    // 1. hash the value, and find the bucket
    // 2. look for a matching "value" in the bucket
    //    remember the "local_iterator" if found
    // 3. increment the size
    // 4. if load_factor() > max_load_factor()
    //     rehash -- the bucket count should grow by 2 * n + 1
    //     determine the new bucket for the value -- DO NOT RECOMPUTE THE HASH
    // 5. emplace a new entry (hash-value)
    //    - at the end of the bucket if no value was found earlier
    //    - before the "local_iterator" remembered earlier (if duplicate existed)
    // 6. use the iterator returned from emplace to make a hash table iterator
    // 7. return iterator

  }

  iterator
  erase (iterator pos)
  {
    // [8] TODO
    //
    // 1. extract the bucket and local_iterator from the iterator
    // 2. call erase of the local_iterator in the bucket's list
    // 3. decrease the size
    // 4. create an iterator pointing to "next"

  }

  void
  rehash (size_t buckets)
  {
    // [9] TODO
    //
    // 1. if size is zero, do nothing
    // 2. create a Bucket to store all of the data
    // 3. walk through all of the buckets and splice them into the bucket created in (2)
    // 4. if "buckets" (param) is non-zero, resize m_buckets
    // 5. walk through each entry in the Bucket created in (2) but leave out the "update"
    //    - splice the current entry into the correct bucket
    //    - NOTE: you will want to store the "next" iterator before calling splice, then
    //        you can update the current iterator to "next"

  }

  // -------------------- Accessors --------------------

  size_t
  size () const noexcept
  {
    return m_size;
  }

  iterator
  find (Value const& key) const
  {
    // [5] TODO
    //
    // 1. compute the hash and determine the location
    // 2. walk throw the bucket with an iterator for-loop
    //    if the hashes match, ensure the key equality also returns true
    //       if it does, construct an iterator -- this, the location, and the current "local" iterator
    // 3. walking through to the end means we didn't find it... return end

  }

  size_t
  bucket (Value const& key) const noexcept
  {
    return m_hash (key) % bucket_count ();
  }

  size_t
  bucket_count () const noexcept
  {
    return m_buckets.size ();
  }

  size_t
  bucket_size (size_t n) const noexcept
  {
    return m_buckets[n].size ();
  }

  float&
  max_load_factor () noexcept
  {
    return m_load_factor;
  }

  float
  max_load_factor () const noexcept
  {
    return m_load_factor;
  }

  float
  load_factor () const noexcept
  {
    return static_cast<float> (static_cast<double> (size ()) / bucket_count ());
  }

  Hash const&
  hash_function () const noexcept
  {
    return m_hash;
  }

  KeyEqual const&
  key_eq () const noexcept
  {
    return m_key_equal;
  }

private:
  Table m_buckets;
  Hash const m_hash;
  KeyEqual const m_key_equal;
  size_t m_size;
  float m_load_factor{DEFAULT_LOAD_FACTOR};
};

#endif
