#ifndef UNORDERED_SET_HPP_
#define UNORDERED_SET_HPP_

#include <cmath>
#include <functional>
#include <limits>
#include <utility>

#include "HashTable.hpp"
#include "support/concepts.hpp"

template<typename Key, typename Hash = std::hash<Key>,
         typename KeyEqual = std::equal_to<Key>>
class UnorderedSet
{
  using table = HashTable<Key, Hash, KeyEqual>;

public:
  using key_type = Key;
  using value_type = Key;
  using size_type = std::size_t;
  using difference_type = std::ptrdiff_t;
  using hasher = Hash;
  using key_equal = KeyEqual;
  using reference = value_type&;
  using const_reference = value_type const&;
  using pointer = value_type*;
  using const_pointer = value_type const*;

  using iterator = typename table::iterator;
  using const_iterator = iterator;
  using local_iterator = typename table::local_iterator;
  using const_local_iterator = local_iterator;

private:
  // -------------------- Constructors --------------------

public:
  UnorderedSet (size_type bucket_count = table::DEFAULT_BUCKET_COUNT,
                hasher const& hash = hasher (),
                key_equal const& equal = key_equal ())
    : m_table (bucket_count, hash, equal)
  {
  }

  template<typename InputIt, Requires (concepts::ForwardIterator, InputIt)>
  UnorderedSet (InputIt first, InputIt last,
                size_type bucket_count = table::DEFAULT_BUCKET_COUNT,
                hasher const& hash = hasher (),
                key_equal const& equal = key_equal ())
    : UnorderedSet (bucket_count, hash, equal)
  {
    insert (first, last);
  }

  UnorderedSet (std::initializer_list<value_type> init,
                size_type bucket_count = table::DEFAULT_BUCKET_COUNT,
                hasher const& hash = hasher (),
                key_equal const& equal = key_equal ())
    : UnorderedSet (init.begin (), init.end (), bucket_count, hash, equal)
  {
  }

  UnorderedSet (UnorderedSet const&) = default;

  UnorderedSet (UnorderedSet&&) = default;

  ~UnorderedSet () = default;

  // -------------------- Assignment --------------------

  UnorderedSet&
  operator= (UnorderedSet const&) = default;

  UnorderedSet&
  operator= (UnorderedSet&&) = default;

  UnorderedSet&
  operator= (std::initializer_list<value_type> ilist);

  // -------------------- Iterators --------------------

  iterator
  begin () noexcept
  {
    return m_table.begin ();
  }

  const_iterator
  begin () const noexcept
  {
    return m_table.begin ();
  }

  const_iterator
  cbegin () const noexcept
  {
    return m_table.cbegin ();
  }

  iterator
  end () noexcept
  {
    return m_table.end ();
  }

  const_iterator
  end () const noexcept
  {
    return m_table.end ();
  }

  const_iterator
  cend () const noexcept
  {
    return m_table.cend ();
  }

  // -------------------- Capacity --------------------
  bool
  empty () const noexcept
  {
    return size () == 0;
  }

  size_type
  size () const noexcept
  {
    return m_table.size ();
  }

  size_type
  max_size () const noexcept
  {
    return std::numeric_limits<size_type>::max ();
  }

  // -------------------- Modifiers --------------------

  void
  clear ()
  {
    m_table.clear ();
  }

  std::pair<iterator, bool>
  insert (value_type const& value)
  {
    return m_table.insert_unique (value);
  }

  std::pair<iterator, bool>
  insert (value_type&& value)
  {
    return m_table.insert_unique (std::move (value));
  }

  template<typename InputIt, Requires (concepts::ForwardIterator, InputIt)>
  void
  insert (InputIt first, InputIt last)
  {
    // [4] TODO

  }

  void
  insert (std::initializer_list<value_type> ilist)
  {
    insert (ilist.begin (), ilist.end ());
  }

  iterator
  erase (const_iterator pos)
  {
    return m_table.erase (pos);
  }

  iterator
  erase (const_iterator first, const_iterator last)
  {
    // [4] TODO

  }

  size_type
  erase (key_type const& key)
  {
    auto pos = find (key);
    if (pos != end ())
    {
      erase (pos);
      return 1;
    }
    return 0;
  }

  // -------------------- Lookup --------------------

  size_type
  count (Key const& key) const
  {
    return static_cast<size_type> (find (key) != end ());
  }

  const_iterator
  find (Key const& key) const
  {
    return m_table.find (key);
  }

  bool
  contains (Key const& key) const
  {
    return find (key) != end ();
  }

  // -------------------- Bucket Interface --------------------

  local_iterator
  begin (size_type n)
  {
    return m_table.begin (n);
  }

  const_local_iterator
  begin (size_type n) const
  {
    return m_table.begin (n);
  }

  const_local_iterator
  cbegin (size_type n) const
  {
    return m_table.begin (n);
  }

  local_iterator
  end (size_type n)
  {
    return m_table.end (n);
  }

  const_local_iterator
  end (size_type n) const
  {
    return m_table.end (n);
  }

  const_local_iterator
  cend (size_type n) const
  {
    return m_table.end (n);
  }

  size_type
  bucket_count () const
  {
    return m_table.bucket_count ();
  }

  size_type
  max_bucket_count () const
  {
    return (size_type (1) << 31) - 1;
  }

  size_type
  bucket_size (size_type n) const
  {
    return m_table.bucket_size (n);
  }

  size_type
  bucket (Key const& key) const
  {
    return m_table.bucket (key);
  }

  // -------------------- Hash Policy --------------------

  float
  load_factor () const
  {
    return m_table.load_factor ();
  }

  float
  max_load_factor () const
  {
    return m_table.max_load_factor ();
  }

  float&
  max_load_factor ()
  {
    return m_table.max_load_factor ();
  }

  void
  rehash (size_type count)
  {
    if (count == bucket_count ())
    {
      return;
    }
    if (count < size () / max_load_factor ())
    {
      count = std::ceil (size () / max_load_factor ());
    }
    m_table.rehash (count);
  }

  void
  reserve (size_type n)
  {
    m_table.rehash (n / max_load_factor ());
  }

  // -------------------- Observers --------------------

  hasher
  hash_function () const
  {
    return m_table.hash_function ();
  }

  key_equal
  key_eq () const
  {
    return m_table.key_eq ();
  }

private:
  table m_table;
};

#endif
