#include "common.hpp"

#include "catch_config.hpp"

#include "catch.hpp"

#include "../HashTable.hpp"

#include <functional>
#include <iomanip>
#include <iterator>

template<typename T>
using CounterHashTable = HashTable<Counter<T>, CounterHash<T>, CounterEq<T>>;

template<typename T, typename Hash, typename Eq>
struct HashTableView
{
  static constexpr float DEFAULT_LOAD_FACTOR = 0.75f;
  static constexpr size_t DEFAULT_BUCKET_COUNT = 13;
  using Table = std::vector<std::list<HashNode<T>>>;
  Table buckets;
  Hash const hash;
  Eq const eq;
  size_t size;
  float load_factor;
};

template<typename T, typename Hash, typename Eq>
auto
inspect (HashTable<T, Hash, Eq>& ref) noexcept
{
  using Ret = HashTableView<T, Hash, Eq>;
  return reinterpret_cast<Ret*> (&ref);
}

template<typename T, typename Hash, typename Eq>
auto
inspect (HashTable<T, Hash, Eq> const& ref) noexcept
{
  using Ret = HashTableView<T, Hash, Eq> const;
  return reinterpret_cast<Ret*> (&ref);
}

template<typename T>
std::ostream&
operator<< (std::ostream& os, HashNode<T> const& node)
{
  os << "( " << node.data << " [..." << std::hex << std::setfill ('0')
     << std::setw (4) << (node.hash & 0xFFFF) << std::dec << "] )";
  return os;
}

template<typename T>
std::ostream&
operator<< (std::ostream& os, std::list<HashNode<T>> const& bucket)
{
  if (bucket.empty ())
    return os;
  os << bucket.front ();
  for (auto i = bucket.begin (); ++i != bucket.end ();)
  {
    os << " -> " << *i;
  }
  return os;
}

template<typename T, typename Hash, typename Eq>
std::ostream&
operator<< (std::ostream& os, HashTable<T, Hash, Eq> const& ref)
{
  auto& buckets = inspect (ref)->buckets;
  os << '\n';
  for (size_t i = 0; i < buckets.size (); ++i)
  {
    os << '[' << i << "] = " << buckets[i] << '\n';
  }
  return os;
}

template<typename T, typename Hash, typename Eq>
std::ostream&
operator<< (std::ostream& os, HashTableIterator<T, Hash, Eq> const& it)
{
  if (it.hash_table () == nullptr)
  {
    os << "unknown @ [" << it.bucket () << "]";
  }
  else
  {

    size_t dist;
    if (it.bucket () == it.hash_table ()->bucket_count ())
    {
      os << "END OF HASHTABLE";
    }
    else
    {
      if (it.local () == it.hash_table ()->end (it.bucket ()))
      {
        os << "end of local";
        dist = 0;
      }
      else
      {
        os << *it;
        dist =
          std::distance (it.hash_table ()->begin (it.bucket ()), it.local ());
      }
      os << " @ [" << it.bucket () << ";" << dist << "]";
    }
  }
  return os;
}

template<typename T, typename Hash, typename Eq>
void
REQUIRE_EXISTS (HashTable<T, Hash, Eq> const& table, T const& val)
{
  size_t bucket_for_val = table.bucket (val);
  CAPTURE (val, bucket_for_val, table);
  REQUIRE (table.size () > 0);
  REQUIRE (table.bucket_size (bucket_for_val) > 0);
  auto& bucket = inspect (table)->buckets[bucket_for_val];
  for (auto i = bucket.begin (); i != bucket.end (); ++i)
  {
    if (table.key_eq () (val, i->data))
    {
      return;
    }
  }
  FAIL ("Could not find value in correct bucket");
}

std::vector<Counter<std::string>> vals{
  Counter<std::string>{"alpha"},   Counter<std::string>{"bravo"},
  Counter<std::string>{"charlie"}, Counter<std::string>{"delta"},
  Counter<std::string>{"echo"},    Counter<std::string>{"foxtrot"},
  Counter<std::string>{"golf"},    Counter<std::string>{"hotel"},
  Counter<std::string>{"india"},   Counter<std::string>{"juliet"},
  Counter<std::string>{"kilo"},    Counter<std::string>{"lima"},
  Counter<std::string>{"mike"},    Counter<std::string>{"november"},
  Counter<std::string>{"oscar"},   Counter<std::string>{"papa"},
  Counter<std::string>{"quebec"},  Counter<std::string>{"romeo"},
  Counter<std::string>{"sierra"},  Counter<std::string>{"tango"},
  Counter<std::string>{"uniform"}, Counter<std::string>{"victor"},
  Counter<std::string>{"whiskey"}, Counter<std::string>{"x-ray"},
  Counter<std::string>{"yankee"},  Counter<std::string>{"zulu"}};

SCENARIO ("HashTable has begin() and end()")
{
  HashTable<int, std::hash<int>, std::equal_to<int>> table;

  int val = 16;
  const size_t hash = table.hash_function () (val);
  const size_t BUCKET = hash % table.bucket_count ();
  inspect (table)->buckets[BUCKET].emplace_back (hash, val);
  inspect (table)->size++;
  WHEN ("We get begin()")
  {
    auto begin = table.begin ();
    THEN ("[1] it represents the \"beginning\" of the table")
    {
      REQUIRE (begin.hash_table () == &table);
      REQUIRE (begin.bucket () == BUCKET);
      REQUIRE (begin.local () == table.begin (BUCKET));
    }
  }
  WHEN ("We get end()")
  {
    auto end = table.end ();
    THEN ("[1] it represents the \"end\" of the table")
    {
      REQUIRE (end.hash_table () == &table);
      REQUIRE (end.bucket () == table.bucket_count ());
    }
  }
}

SCENARIO ("HashTableIterator can be compared", "[HashTableIterator][comp]")
{
  using iter = HashTableIterator<int, std::hash<int>, std::equal_to<int>>;
  using table = HashTable<int, std::hash<int>, std::equal_to<int>>;

  table s, t;
  WHEN ("Two iterators are the same")
  {
    THEN ("[1] they are deemed equal")
    {
      iter i (&s, 0, s.begin (0));
      iter j (&s, 0, s.begin (0));
      REQUIRE (i == j);
    }
  }
  WHEN ("Two iterators differ by table")
  {
    THEN ("[1] they are not equal")
    {
      iter i (&s, 0, s.begin (0));
      iter j (&t, 0, t.begin (0));
      REQUIRE_FALSE (i == j);
    }
  }
  WHEN ("Two iterators differ by bucket")
  {
    THEN ("[1] they are not equal")
    {
      iter i (&s, 0, s.begin (0));
      iter j (&s, 1, s.begin (1));
      REQUIRE_FALSE (i == j);
    }
  }
  WHEN ("Two iterators differ only by local_iterator (except for end())")
  {
    THEN ("[1] they are not equal")
    {
      inspect (s)->buckets[0].emplace_back (0, 0);
      iter i (&s, 0, s.begin (0));
      iter j (&s, 0, s.end (0));
      REQUIRE_FALSE (i == j);
    }
  }
}

SCENARIO ("HashTableIterator can be incremented", "[HashTableIterator][incr]")
{
  GIVEN ("An arbitrary HashTable")
  {
    HashTable<int, std::hash<int>, std::equal_to<int>> table;
    auto& hash = table.hash_function ();
    auto& t = *inspect (table);
    std::initializer_list<int> values = {26, 13, 4, 18, 34, 8, 21, 25};
    for (int v : values)
    {
      t.buckets[table.bucket (v)].emplace_back (hash (v), v);
    }
    t.size += 8;
    std::set<int> set (values.begin (), values.end ());
    THEN ("[4] increment should work")
    {
      auto prev = table.begin ();
      for (auto i = std::next (prev); i != table.end (); prev = i++)
      {
        if (prev.bucket () == i.bucket ())
        {
          INFO (
            "Advancing an iterator which has information in the same bucket "
            "should just advance the local");
          REQUIRE (std::next (prev.local ()) == i.local ());
        }
        else
        {
          INFO ("Advancing an iterator should go to the next biggest bucket");
          REQUIRE (prev.bucket () < i.bucket ());
        }
      }
      size_t count = 0;
      for (auto i = table.begin (); i != table.end (); ++i)
      {
        count += set.erase (*i);
      }
      {
        INFO ("All elements should be traversed");
        REQUIRE (set.empty ());
        REQUIRE (count == table.size ());
      }
    }
  }
}

SCENARIO ("We can insert_unique into a HashTable", "[HashTable][insert_unique]")
{
  GIVEN ("An empty HashTable")
  {
    CounterHashTable<std::string> table;
    WHEN ("We insert a value")
    {
      Counter<std::string> val1 ("alice");
      Counters::reset ();
      auto retVal = table.insert_unique (val1);
      THEN ("[2] It is added to the HashTable")
      {
        {
          INFO ("hash should only be computed once!");
          REQUIRE (Counters::fn_hash == 1);
        }
        INFO ("Table after insert:\n" << table);
        {
          INFO ("iterator should refer to value inserted");
          REQUIRE (val1 == *(retVal.first));
        }
        {
          INFO ("second part of pair should be true");
          REQUIRE (retVal.second);
        }
        {
          INFO ("size should increase");
          REQUIRE (table.size () == 1);
        }
        {
          INFO ("should have been placed into the correct bucket");
          REQUIRE (table.bucket_size (table.bucket (val1)) == 1);
        }
      }
      AND_WHEN ("We insert the same value")
      {
        auto retVal = table.insert_unique (val1);
        THEN ("[2] There is only one value in the hash table")
        {
          INFO ("Table after repeated insert:\n" << table);
          {
            INFO ("iterator should refer to value inserted");
            REQUIRE (val1 == *(retVal.first));
          }
          {
            INFO ("second part of pair should be true");
            REQUIRE_FALSE (retVal.second);
          }
          REQUIRE (table.size () == 1);
        }
      }
    }
    WHEN ("We insert a bunch of values")
    {
      size_t limit = size_t (13 * 0.75f);
      for (size_t i = 0; i < limit; ++i)
      {
        table.insert_unique (vals[i]);
      }
      THEN ("[2] They are all added to the HashTable")
      {
        for (size_t i = 0; i < limit; ++i)
        {
          REQUIRE_EXISTS (table, vals[i]);
        }
        {
          INFO ("size was updated");
          REQUIRE (table.size () == limit);
        }
      }
      WHEN ("We insert more values which exceeds the load factor")
      {
        size_t old_bucket_count = table.bucket_count ();
        table.insert_unique (vals[limit++]);
        THEN ("[2] rehashing occurs")
        {
          REQUIRE (old_bucket_count != table.bucket_count ());
        }
        size_t new_limit = (27 * 0.75f);
        for (size_t i = limit; i < new_limit; ++i)
        {
          table.insert_unique (vals[i]);
        }
        THEN ("[1] they are added to the HashTable")
        {
          for (size_t i = 0; i < new_limit; ++i)
          {
            REQUIRE_EXISTS (table, vals[i]);
          }
        }
      }
    }
  }
}

SCENARIO ("We can insert into a HashTable", "[HashTable][insert]")
{
  GIVEN ("An empty HashTable")
  {
    CounterHashTable<std::string> table;
    WHEN ("We insert a value")
    {
      Counter<std::string> val1 ("alice");
      Counters::reset ();
      auto retVal = table.insert (val1);
      THEN ("[2] It is added to the HashTable")
      {
        {
          INFO ("hash should only be computed once!");
          REQUIRE (Counters::fn_hash == 1);
        }
        INFO ("Table after insert:\n" << table);
        {
          INFO ("iterator should refer to value inserted");
          REQUIRE (val1 == *retVal);
        }
        {
          INFO ("size should increase");
          REQUIRE (table.size () == 1);
        }
        {
          INFO ("should have been placed into the correct bucket");
          REQUIRE (table.bucket_size (table.bucket (val1)) == 1);
        }
      }
      AND_WHEN ("We insert the same value")
      {
        auto retVal = table.insert (val1);
        THEN ("[2] There are now two values")
        {
          INFO ("Table after repeated insert:\n" << table);
          {
            INFO ("iterator should refer to value inserted");
            REQUIRE (val1 == *retVal);
          }
          REQUIRE (table.size () == 2);
        }
      }
      AND_WHEN ("We insert a bunch of values")
      {
        for (int i = 1; i < 13; ++i)
        {
          if (i == 7 || i == 10)
          {
            continue;
          }
          table.insert (vals[i]);
        }

        auto i1 = table.insert (vals[7]);
        auto j1 = table.insert (vals[10]);
        auto i2 = table.insert (vals[7]);
        auto j2 = table.insert (vals[10]);
        auto i3 = table.insert (vals[7]);
        auto j3 = table.insert (vals[10]);

        CAPTURE (i1, i2, i3, j1, j2, j3, table);

        THEN ("[1] they are added to the HashTable")
        {
          REQUIRE (*i1 == vals[7]);
          REQUIRE (*i2 == vals[7]);
          REQUIRE (*i3 == vals[7]);
          REQUIRE (*j1 == vals[10]);
          REQUIRE (*j2 == vals[10]);
          REQUIRE (*j2 == vals[10]);
        }
        THEN ("[2] they are all separate values")
        {
          REQUIRE (i1 != i2);
          REQUIRE (i2 != i3);
          REQUIRE (i3 != i1);
          REQUIRE (j1 != j2);
          REQUIRE (j2 != j3);
          REQUIRE (j3 != j1);
        }
        THEN ("[2] all values of the same value are grouped together")
        {
          REQUIRE (std::next (i3) == i2);
          REQUIRE (std::next (i2) == i1);
          REQUIRE (std::next (j3) == j2);
          REQUIRE (std::next (j2) == j1);
        }
      }
    }
  }
}

SCENARIO ("HashTable can be cleared", "[HashTable][clear]")
{
  GIVEN ("A hash table with a bunch of values")
  {
    CounterHashTable<std::string> table;
    size_t size = 0;
    for (int i = 0; i < 3; ++i)
    {
      for (auto& val : vals)
      {
        table.insert (val);
        ++size;
      }
    }
    REQUIRE (table.size () == size);
    std::vector<size_t> sizes (table.bucket_count ());
    for (size_t i = 0; i < table.bucket_count (); ++i)
    {
      sizes[i] = table.bucket_size (i);
    }
    WHEN ("We call clear")
    {
      table.clear ();
      THEN ("[1] size becomes zero")
      {
        REQUIRE (table.size () == 0);
      }
      THEN ("[1] bucket count remains the same")
      {
        REQUIRE (table.bucket_count () == sizes.size ());
      }
      THEN ("[2] each bucket has been cleared")
      {
        for (size_t i = 0; i < table.bucket_count (); ++i)
        {
          REQUIRE (table.bucket_size (i) == 0);
        }
      }
    }
  }
}

// find [5]
SCENARIO ("We should be able to find values", "[HashTable][find]")
{
  GIVEN ("A populated Hash Table")
  {
    HashTable<int, std::hash<int>, std::equal_to<int>> table;
    for (int i = 0; i < 100; i += 2)
    {
      table.insert (i);
    }
    WHEN ("We look for a value which doesn't exist")
    {
      auto it = table.find (1);
      THEN ("[1] we return end")
      {
        REQUIRE (it == table.end ());
      }
    }
    WHEN ("We look for a value which does exist")
    {
      auto it = table.find (50);
      THEN ("[2] we return an iterator pointing to it")
      {
        REQUIRE (*it == 50);
      }
      WHEN ("We look for a value which has a duplicate")
      {
        auto jt = table.insert (50);
        auto kt = table.insert (50);
        auto find_result = table.find (50);
        THEN ("[2] we return the first instance of the value")
        {
          CAPTURE (it, jt, kt, find_result);
          REQUIRE (*find_result == 50);
          REQUIRE (
            (find_result == it || find_result == jt || find_result == kt));
          REQUIRE (*++find_result == 50);
          REQUIRE (
            (find_result == it || find_result == jt || find_result == kt));
          REQUIRE (*++find_result == 50);
          REQUIRE (
            (find_result == it || find_result == jt || find_result == kt));
        }
      }
    }
  }
}

// erase [8]
SCENARIO ("We should be able to erase from a HashTable", "[HashTable][erase]")
{
  GIVEN ("A populated Hash Table")
  {
    CounterHashTable<int> table;
    table.max_load_factor () = 10000.0f;
    for (int i = 0; i < 100; ++i)
    {
      table.insert (Counter<int> (i));
    }
    size_t size = 100;
    Counter<int> val;
    Counters::reset ();
    WHEN ("We attempt to remove any value")
    {
      THEN ("[8] erase functions as expected")
      {
        for (int i = 0; i < 100; ++i)
        {
          val.data = i;
          size_t bucket;
          size_t bucket_size;
          auto loc = table.find (val);
          {
            INFO ("we should always find the value to remove");
            REQUIRE (loc != table.end ());
          }
          auto next = std::next (loc);
          bucket = loc.bucket ();
          bucket_size = inspect (table)->buckets[bucket].size ();
          auto retVal = table.erase (loc);
          {
            INFO ("calling erase should return the \"next\" iterator");
            REQUIRE (next == retVal);
          }
          {
            INFO ("size should decrease");
            REQUIRE (table.size () == size - 1);
          }
          {
            INFO ("bucket size should have decreased");
            size_t new_bucket_size = inspect (table)->buckets[bucket].size ();
            REQUIRE (new_bucket_size == bucket_size - 1);
          }
          --size;
          {
            INFO ("The element should have been destroyed")
            {
              REQUIRE (Counters::dtor == 1);
              Counters::dtor = 0;
            }
          }
        }
      }
    }
  }
}

// rehash [9]
SCENARIO ("We should be able to rehash a HashTable", "[HashTable][rehash]")
{
  GIVEN ("A populated Hash Table")
  {
    std::set<Counter<int>> set;
    CounterHashTable<int> table;
    std::vector<std::list<int>::iterator> iters;
    std::vector<typename CounterHashTable<int>::iterator> table_iters;
    std::vector<typename CounterHashTable<int>::iterator> new_table_iters;
    table.max_load_factor () = 10000.0f;
    size_t SIZE = 20;
    Counter<int> val;
    for (size_t i = 0; i < SIZE; ++i)
    {
      val.data = i;
      table.insert (val);
      set.insert (val);
    }
    size_t old_bucket_count = table.bucket_count ();
    for (auto i = table.begin (); i != table.end (); ++i)
    {
      table_iters.emplace_back (i);
    }
    Counters::reset ();

    WHEN ("We rehash(0)")
    {
      table.rehash (0);
      {
        INFO ("We should never rehash; use the hash stored in the \"node\"");
        REQUIRE (Counters::fn_hash == 0);
      }

      THEN ("[1] Nothing should happen")
      {
        for (auto i = table.begin (); i != table.end (); ++i)
        {
          new_table_iters.emplace_back (i);
        }
        {
          INFO ("Bucket count should not change");
          REQUIRE (table.bucket_count () == old_bucket_count);
        }
        {
          INFO ("All iterators should stay the same");
          auto old = table_iters.begin ();
          auto cur = new_table_iters.begin ();
          while (old != table_iters.end ())
          {
            REQUIRE (*old == *cur);
            ++old;
            ++cur;
          }
        }
      }
    }
    WHEN ("We rehash to a smaller bucket count")
    {
      table.rehash (old_bucket_count / 2);
      {
        INFO ("We should never rehash; use the hash stored in the \"node\"");
        REQUIRE (Counters::fn_hash == 0);
      }

      THEN ("[4] elements should be spliced in and out")
      {
        {
          INFO ("Bucket count should change");
          REQUIRE (table.bucket_count () == old_bucket_count / 2);
        }
        {
          INFO ("All hashes should be placed in the correct bucket");
          auto const& buckets = inspect(table)->buckets;
          for (size_t bucket_idx = 0; bucket_idx < buckets.size(); ++bucket_idx)
          {
            size_t local_idx = 0;
            for (auto& [hash, data] : buckets[bucket_idx])
            {
              CAPTURE (bucket_idx, local_idx, hash, data);
              REQUIRE (hash % buckets.size() == bucket_idx);
              ++local_idx;
            }
          }
        }
        {
          INFO ("If this segfaults, then you are not splicing correctly");
          for (auto i = table_iters.begin (); i != table_iters.end (); ++i)
          {
            INFO (
              "We can still use the local iter to access the original values");
            INFO ("If we cannot, then splicing has majestically failed");
            REQUIRE (1 == set.erase (i->local ()->data));
          }
          REQUIRE (set.empty ());
        }
      }
    }
    WHEN ("We rehash to a larger bucket count")
    {
      table.rehash (old_bucket_count * 2 + 1);
      {
        INFO ("We should never rehash; use the hash stored in the \"node\"");
        REQUIRE (Counters::fn_hash == 0);
      }

      THEN ("[4] elements should be spliced in and out")
      {
        {
          INFO ("Bucket count should change");
          REQUIRE (table.bucket_count () == old_bucket_count * 2 + 1);
        }
        {
          INFO ("All hashes should be placed in the correct bucket");
          auto const& buckets = inspect(table)->buckets;
          for (size_t bucket_idx = 0; bucket_idx < buckets.size(); ++bucket_idx)
          {
            size_t local_idx = 0;
            for (auto& [hash, data] : buckets[bucket_idx])
            {
              CAPTURE (bucket_idx, local_idx, hash, data);
              REQUIRE (hash % buckets.size() == bucket_idx);
              ++local_idx;
            }
          }
        }
        {
          INFO ("If this segfaults, then you are not splicing correctly");
          for (auto i = table_iters.begin (); i != table_iters.end (); ++i)
          {
            INFO (
              "We can still use the local iter to access the original values");
            INFO ("If we cannot, then splicing has majestically failed");
            REQUIRE (1 == set.erase (i->local ()->data));
          }
          REQUIRE (set.empty ());
        }
      }
    }
  }
}
