#ifndef CONCEPTS_HPP_
#define CONCEPTS_HPP_

#include <experimental/type_traits>

namespace concepts
{

template<typename T>
auto
val () -> decltype (std::declval<T> ())
{
  return std::declval<T> ();
}

template<typename T>
auto
cval () -> decltype (std::declval<T const> ())
{
  return std::declval<T const> ();
}

template<typename T>
auto
ref () -> decltype (std::declval<T &> ())
{
  return std::declval<T &> ();
}
template<typename T>
auto
cref () -> decltype (std::declval<T const &> ())
{
  return std::declval<T const &> ();
}
template<typename... T>
constexpr std::true_type
valid_expr (T &&...);

template<typename Ret, typename T>
constexpr Ret
returns (T const &);

template<typename T, typename U>
constexpr auto
convertible_to (U &&u)
  -> decltype (returns<std::true_type> (static_cast<T> ((U &&) u)));

namespace impl
{

template<typename...>
struct list;

template<bool B>
using bool_ = std::integral_constant<bool, B>;

template<typename, typename = std::true_type>
struct if_
{
};

template<typename If>
struct if_<list<If>, bool_<If::type::value>>
{
  using type = void;
};

template<typename If, typename Then>
struct if_<list<If, Then>, bool_<If::type::value>>
{
  using type = Then;
};

template<typename If, typename Then, typename Else>
struct if_<list<If, Then, Else>, bool_<If::type::value>>
{
  using type = Then;
};

template<typename If, typename Then, typename Else>
struct if_<list<If, Then, Else>, bool_<!If::type::value>>
{
  using type = Else;
};

template<bool...>
struct blist;

template<typename T>
using negate_t = bool_<!T::value>;

template<bool... Bs>
using all_of = std::is_same<blist<true, Bs...>, blist<Bs..., true>>;

template<bool... Bs>
using none_of = std::is_same<blist<false, Bs...>, blist<Bs..., false>>;

template<bool... Bs>
using any_of = negate_t<none_of<Bs...>>;

template<typename... Bs>
using all_of_t = all_of<Bs::value...>;

template<typename... Bs>
using none_of_t = none_of<Bs::value...>;

template<typename... Bs>
using any_of_t = any_of<Bs::value...>;

} // namespace impl

template<typename... Ts>
using if_ = typename impl::if_<impl::list<Ts...>>::type;

template<bool If, typename... Args>
using if_c = typename impl::if_<impl::list<impl::bool_<If>, Args...>>::type;

template<typename T, typename U>
constexpr auto has_type (U &&) -> if_<std::is_same<T, U>, std::true_type>;

template<typename T>
auto
models () -> if_<std::is_same<std::true_type, T>, std::true_type>;

template<template<typename...> class Thing, typename... Args>
using requires_ = std::experimental::is_detected<Thing, Args...>;

template<typename... Args>
using all_of = impl::all_of_t<Args...>;

template<typename... Args>
using none_of = impl::none_of_t<Args...>;

template<typename... Args>
using any_of = impl::any_of_t<Args...>;
}

#define DefineConcept(...) decltype (concepts::valid_expr (__VA_ARGS__))

#define Requires(ConceptName,...) \
  std::enable_if_t<::concepts::requires_<ConceptName, __VA_ARGS__>::value>* = nullptr

namespace concepts
{

template<typename T>
using LessThanComparable = DefineConcept (convertible_to<bool> (val<T> () < val<T> ()));

template<typename T>
using GreaterThanComparable = DefineConcept (convertible_to<bool> (val<T> () > val<T> ()));

template<typename T>
using LessEqualComparable = DefineConcept (convertible_to<bool> (val<T> () <= val<T> ()));

template<typename T>
using GreaterEqualComparable = DefineConcept (convertible_to<bool> (val<T> () >= val<T> ()));

template<typename T>
using EqualityComparable = DefineConcept (convertible_to<bool> (val<T> () == val<T> ()));

template<typename T, typename U>
using ComparableTo =
  DefineConcept (convertible_to<bool> (val<U> () < val<T> ()),
                 convertible_to<bool> (val<T> () < val<U> ()),
                 convertible_to<bool> (val<U> () <= val<T> ()),
                 convertible_to<bool> (val<T> () <= val<U> ()),
                 convertible_to<bool> (val<U> () > val<T> ()),
                 convertible_to<bool> (val<T> () > val<U> ()),
                 convertible_to<bool> (val<U> () >= val<T> ()),
                 convertible_to<bool> (val<T> () >= val<U> ()),
                 convertible_to<bool> (val<U> () == val<T> ()),
                 convertible_to<bool> (val<T> () == val<U> ()),
                 convertible_to<bool> (val<U> () != val<T> ()),
                 convertible_to<bool> (val<T> () != val<U> ()));

template<typename T>
using Comparable = ComparableTo<T, T>;

template<typename T>
using Iterator = DefineConcept (*val<T> (),
                                has_type<T &> (++ref<T> ()),
                                models<EqualityComparable<T>> ());

template<typename T>
using ForwardIterator = DefineConcept (models<Iterator<T>> (),
                                       ref<T> ()++,
                                       *ref<T> ()++);

template<typename T>
using BidirectionalIterator =
  DefineConcept (models<ForwardIterator<T>> (),
                 has_type<T &> (--ref<T> ()),
                 convertible_to<T const &> (ref<T> ()--),
                 *ref<T> ()--);

template<typename T>
using diff_t = decltype (val<T> () - val<T> ());

template<typename T>
using RandomAccessIterator = DefineConcept (
  models<BidirectionalIterator<T>> (),
  models<Comparable<T>> (),
  has_type<T &> (ref<T> () += val<diff_t<T>> ()),
  has_type<T> (val<T> () + val<diff_t<T>> ()),
  has_type<T> (val<diff_t<T>> () + val<T> ()),
  has_type<T &> (ref<T> () -= val<diff_t<T>> ()),
  has_type<T> (val<T> () - val<diff_t<T>> ()),
  val<T> ()[val<diff_t<T>> ()]);
}

#endif
